import pytest

from psycopg import errors as e
from psycopg import pq
from psycopg.adapt import Transformer

try:
    from psycopg._copy_base import parse_row_binary
except ImportError:
    # Allow importing this test module (not to run it) with an old psycopg
    # imjplementation (to run the pool tests)
    pass


def test_copy_binary_parse_length_exceeding_data():
    tx = Transformer()
    # One field declaring length 10, but only two bytes ("ab") present.
    data = bytes.fromhex("00010000000a6162")
    with pytest.raises(e.DataError, match="length exceeding data"):
        parse_row_binary(data, tx)


def test_copy_binary_parse_exact_length():
    tx = Transformer()
    tx.set_loader_types([25], pq.Format.TEXT)  # oid 25 = text
    # One field declaring length 2, with exactly two bytes ("ab") present.
    data = bytes.fromhex("0001000000026162")
    assert parse_row_binary(data, tx) == ("ab",)


@pytest.fixture
def tx2():
    tx = Transformer()
    tx.set_loader_types([25, 25], pq.Format.TEXT)
    return tx


@pytest.mark.parametrize(
    "data",
    [
        pytest.param(b"", id="empty"),
        pytest.param(b"\x00", id="field-count"),
        pytest.param(b"\x00\x01", id="field-length-missing"),
        pytest.param(b"\x00\x01\x00\x00\x00", id="field-length"),
    ],
)
def test_parse_row_binary_truncated_header(tx2, data):
    with pytest.raises(e.DataError, match="bad copy data"):
        parse_row_binary(data, tx2)


def test_parse_row_binary_length_exceeding_data(tx2):
    # One field announcing 4 bytes but only 2 present.
    data = b"\x00\x01" + b"\x00\x00\x00\x04" + b"ab"
    with pytest.raises(e.DataError, match="length exceeding data"):
        parse_row_binary(data, tx2)


def test_parse_row_binary_ok(tx2):
    data = b"\x00\x02" + b"\x00\x00\x00\x02" + b"ab" + b"\xff\xff\xff\xff"
    assert parse_row_binary(data, tx2) == ("ab", None)



"""
Mappings between PostgreSQL and Python encodings.
"""

# Copyright (C) 2020 The Psycopg Team

from __future__ import annotations

import re
import codecs
import string
from typing import TYPE_CHECKING, Any
from functools import cache

from .errors import NotSupportedError

if TYPE_CHECKING:
    from ._connection_base import BaseConnection

# From: https://www.postgresql.org/docs/current/multibyte.html#CHARSET-TABLE
_py_codecs = (
    (("BIG5", "WIN950", "Windows950"), "big5"),
    (("EUC_CN",), "gb2312"),
    (("EUC_JIS_2004",), "euc_jis_2004"),
    (("EUC_JP",), "euc_jp"),
    (("EUC_KR",), "euc_kr"),
    # (("EUC_TW",), not available in Python
    (("GB18030",), "gb18030"),
    (("GBK", "WIN936", "Windows936"), "gbk"),
    (("ISO_8859_5",), "iso8859-5"),
    (("ISO_8859_6",), "iso8859-6"),
    (("ISO_8859_7",), "iso8859-7"),
    (("ISO_8859_8",), "iso8859-8"),
    (("JOHAB",), "johab"),
    (("KOI8R", "KOI8"), "koi8-r"),
    (("KOI8U",), "koi8-u"),
    (("LATIN1", "ISO88591"), "iso8859-1"),
    (("LATIN10", "ISO885916"), "iso8859-16"),
    (("LATIN2", "ISO88592"), "iso8859-2"),
    (("LATIN3", "ISO88593"), "iso8859-3"),
    (("LATIN4", "ISO88594"), "iso8859-4"),
    (("LATIN5", "ISO88599"), "iso8859-9"),
    (("LATIN6", "ISO885910"), "iso8859-10"),
    (("LATIN7", "ISO885913"), "iso8859-13"),
    (("LATIN8", "ISO885914"), "iso8859-14"),
    (("LATIN9", "ISO885915"), "iso8859-15"),
    # "MULE_INTERNAL" not available in Python
    (("SHIFT_JIS_2004",), "shift_jis_2004"),
    (("SJIS", "Mskanji", "ShiftJIS", "WIN932", "Windows932"), "shift_jis"),
    # this actually means no encoding, see PostgreSQL docs
    # it is special-cased by the text loader.
    (("SQL_ASCII",), "ascii"),
    (("UHC", "WIN949", "Windows949"), "cp949"),
    (("UTF8", "Unicode"), "utf-8"),
    (("WIN1250",), "cp1250"),
    (("WIN1251", "WIN"), "cp1251"),
    (("WIN1252",), "cp1252"),
    (("WIN1253",), "cp1253"),
    (("WIN1254",), "cp1254"),
    (("WIN1255",), "cp1255"),
    (("WIN1256",), "cp1256"),
    (("WIN1257",), "cp1257"),
    (("WIN1258", "ABC", "TCVN", "TCVN5712", "VSCII"), "cp1258"),
    (("WIN866", "ALT"), "cp866"),
    (("WIN874",), "cp874"),
)

py_codecs = {a.upper().encode(): v for aliases, v in _py_codecs for a in aliases}

# Add an alias without underscore, for lenient lookups
py_codecs.update(
    (a.replace("_", "").encode(), v)
    for aliases, v in _py_codecs
    for a in aliases
    if "_" in a
)

pg_codecs = {v: k[0].encode() for k, v in _py_codecs}


def conn_encoding(conn: BaseConnection[Any] | None) -> str:
    """
    Return the Python encoding name of a psycopg connection.

    Default to utf8 if the connection has no encoding info.
    """
    return conn.pgconn._encoding if conn else "utf-8"


def conninfo_encoding(conninfo: str) -> str:
    """
    Return the Python encoding name passed in a conninfo string. Default to utf8.

    Because the input is likely to come from the user and not normalised by the
    server, be somewhat lenient (non-case-sensitive lookup, ignore noise chars).
    """
    from .conninfo import conninfo_to_dict

    params = conninfo_to_dict(conninfo)
    if pgenc := params.get("client_encoding"):
        try:
            return pg2pyenc(str(pgenc).encode())
        except NotSupportedError:
            pass

    return "utf-8"


@cache
def py2pgenc(name: str) -> bytes:
    """Convert a Python encoding name to PostgreSQL encoding name.

    Raise LookupError if the Python encoding is unknown.
    """
    return pg_codecs[codecs.lookup(name).name]


@cache
def pg2pyenc(name: bytes) -> str:
    """Convert a PostgreSQL encoding name to Python encoding name.

    Raise NotSupportedError if the PostgreSQL encoding is not supported by
    Python.
    """
    try:
        return py_codecs[name.replace(b"-", b"").replace(b"_", b"").upper()]
    except KeyError:
        sname = name.decode("utf8", "replace")
        raise NotSupportedError(f"codec not available in Python: {sname!r}")


def _as_python_identifier(s: str, prefix: str = "f") -> str:
    """
    Reduce a string to a valid Python identifier.

    Replace all non-valid chars with '_' and prefix the value with `!prefix` if
    the first letter is an '_'.
    """
    if not s.isidentifier():
        if s[0] in "1234567890":
            s = prefix + s
        if not s.isidentifier():
            s = _re_clean.sub("_", s)
    # namedtuple fields cannot start with underscore. So...
    if s[0] == "_":
        s = prefix + s
    return s


_re_clean = re.compile(
    f"[^{string.ascii_lowercase}{string.ascii_uppercase}{string.digits}_]"
)



# WARNING: this file is auto-generated by 'async_to_sync.py'
# from the original file 'test_tstring_async.py'
# DO NOT CHANGE! Change the original file instead.
from random import random
from typing import Any

import pytest

import psycopg
from psycopg import sql
from psycopg.pq import Format

from .acompat import skip_async
from .test_sql import no_e
from .test_adapt import make_dumper

vstr = "hello"
vint = 16


def test_connection_no_params(conn):
    with pytest.raises(TypeError):
        conn.execute(t"select 1", [])


def test_cursor_no_params(conn):
    cur = conn.cursor()
    with pytest.raises(TypeError):
        cur.execute(t"select 1", [])


def test_connection_execute(conn):
    cur = conn.execute(t"select {vstr}")
    assert cur.fetchone() == ("hello",)
    assert cur._query.query == b"select $1"
    assert cur._query.params == [b"hello"]
    assert cur._query.types == (0,)


@pytest.mark.parametrize(
    "t", [t"select {vstr!a}", t"select {vstr!r}", t"select {vstr!s}"]
)
def test_no_conversion(conn, t):
    with pytest.raises(TypeError):
        conn.execute(t)


@pytest.mark.parametrize(
    "t, fmt",
    [
        (t"select {vint}", Format.BINARY),
        (t"select {vint:s}", Format.BINARY),
        (t"select {vint:t}", Format.TEXT),
        (t"select {vint:b}", Format.BINARY),
    ],
)
def test_format(conn, t, fmt):
    cur = conn.execute(t)
    assert cur.fetchone() == (16,)
    assert cur._query.query == b"select $1"
    assert cur._query.types == (psycopg.adapters.types["smallint"].oid,)
    assert cur._query.params == [b"\x00\x10" if fmt == Format.BINARY else b"16"]
    assert cur._query.formats == [fmt]


def test_format_bad(conn):
    with pytest.raises(psycopg.ProgrammingError, match="format 'x' not supported"):
        conn.execute(t"select {vint:x}")


def test_expression(conn):
    cur = conn.execute(t"select {vint * 2}")
    assert cur.fetchone() == (32,)
    assert cur._query.query == b"select $1"
    assert cur._query.types == (psycopg.adapters.types["smallint"].oid,)
    assert cur._query.params == [b"\x00 "]
    assert cur._query.formats == [Format.BINARY]


def test_format_identifier(conn):
    f1 = "foo-bar"
    f2 = "baz"
    cur = conn.execute(t"select {vint} as {f1:i}, {vint * 2:t} as {f2:i}")
    assert cur.fetchone() == (16, 32)
    assert cur._query.query == b'select $1 as "foo-bar", $2 as "baz"'
    assert cur._query.types == (psycopg.adapters.types["smallint"].oid,) * 2
    assert cur._query.params == [b"\x00\x10", b"32"]
    assert cur._query.formats == [Format.BINARY, Format.TEXT]


def test_format_literal(conn):
    f1 = "foo-bar"
    f2 = "baz"
    cur = conn.execute(t"select {vint * 2:l} as {f1:i}, {vint:t} as {f2:i}")
    assert cur.fetchone() == (32, 16)
    assert cur._query.query == b'select 32 as "foo-bar", $1 as "baz"'
    assert cur._query.types == (psycopg.adapters.types["smallint"].oid,)
    assert cur._query.params == [b"16"]
    assert cur._query.formats == [Format.TEXT]


def test_nested(conn):
    part = t"{vint} as foo"
    cur = conn.execute(t"select {part:q}")
    assert cur.fetchone() == (16,)
    assert cur._query.query == b"select $1 as foo"
    assert cur._query.types == (psycopg.adapters.types["smallint"].oid,)
    assert cur._query.params == [b"\x00\x10"]
    assert cur._query.formats == [Format.BINARY]

    with pytest.raises(psycopg.ProgrammingError, match="Template.*':q'"):
        cur = conn.execute(t"select {part}")


def test_scope(conn):
    t = t"select "
    for i, name in enumerate(("foo", "bar", "baz")):
        if i:
            t += t", "
        t += t"{i} as {name:i}"

    cur = conn.execute(t)
    assert cur.fetchone() == (0, 1, 2)
    assert cur.description[0].name == "foo"
    assert cur.description[2].name == "baz"


def test_no_reuse(conn):
    t = t"select {vint}, {vint}"
    cur = conn.execute(t)
    assert cur.fetchone() == (vint, vint)
    assert b"$2" in cur._query.query


def test_volatile(conn):
    t = t"select {random()}, {random()}"
    cur = conn.execute(t)
    rec = cur.fetchone()
    assert rec[0] != rec[1]
    assert b"$2" in cur._query.query


def test_sql(conn):
    part = sql.SQL("foo")
    cur = conn.execute(t"select {vint} as {part:q}")
    assert cur.fetchone() == (16,)
    assert cur._query.query == b"select $1 as foo"

    with pytest.raises(psycopg.ProgrammingError, match="sql\\.SQL.*':q'"):
        conn.execute(t"select {vint} as {part:i}")


def test_sql_composed(conn):
    part = sql.SQL("{} as {}").format(vint, sql.Identifier("foo"))
    cur = conn.execute(t"select {part:q}")
    assert cur.fetchone() == (16,)
    assert cur._query.query == b'select 16 as "foo"'

    with pytest.raises(psycopg.ProgrammingError, match="sql\\.Composed.*':q'"):
        conn.execute(t"select {part}")


def test_sql_identifier(conn):
    part = sql.Identifier("foo")
    cur = conn.execute(t"select {vint} as {part:i}")
    assert cur.fetchone() == (16,)
    assert cur._query.query == b'select $1 as "foo"'

    with pytest.raises(psycopg.ProgrammingError, match="sql\\.Identifier.*':i'"):
        conn.execute(t"select {vint} as {part}")


def test_sql_literal(conn):
    lit = sql.Literal(42)
    cur = conn.execute(t"select {lit:l} as foo")
    assert cur.fetchone() == (42,)
    assert cur._query.query == b"select 42 as foo"

    with pytest.raises(psycopg.ProgrammingError, match="sql\\.Literal.*':l'"):
        conn.execute(t"select {lit} as foo")


def test_sql_placeholder(conn):
    part = sql.Placeholder("foo")
    with pytest.raises(psycopg.ProgrammingError, match="Placeholder not supported"):
        conn.execute(t"select {part}")


@pytest.mark.xfail(reason="Template.join() needed")
def test_template_join(conn):
    ts = [t"{i} as {name:i}" for i, name in enumerate(("foo", "bar", "baz"))]
    fields = t",".join(ts)  # type: ignore[attr-defined]
    cur = conn.execute(t"select {fields}")
    assert cur.fetchone() == (0, 1, 2)
    assert cur.description[0].name == "foo"
    assert cur.description[2].name == "baz"


def test_sql_join(conn):
    ts = [t"{i} as {name:i}" for i, name in enumerate(("foo", "bar", "baz"))]
    fields = sql.SQL(",").join(ts)
    cur = conn.execute(t"select {fields:q}")
    assert cur.fetchone() == (0, 1, 2)
    assert cur.description[0].name == "foo"
    assert cur.description[2].name == "baz"


@skip_async
def test_sql_join_mixed_types():
    ts: list[Any] = [t"{i} as {name:i}" for i, name in enumerate(("foo", "bar"))]
    ts.append(sql.Literal(2))
    for it in (ts, reversed(ts)):
        with pytest.raises(TypeError, match="Template and Literal"):
            sql.SQL(",").join(it)


def test_copy(conn):
    cur = conn.cursor()
    with cur.copy(t"copy (select * from generate_series(1, {3})) to stdout") as copy:
        data = list(copy.rows())
    assert data == [("1",), ("2",), ("3",)]


def test_client_cursor(conn):
    cur = psycopg.ClientCursor(conn)
    cur.execute(t"select {vint}, {vstr} as {vstr:i}")
    assert cur.fetchone() == (vint, vstr)
    assert cur.description
    assert cur.description[1].name == vstr
    assert cur._query
    assert str(vint) in cur._query.query.decode()
    assert str(vint) == cur._query.params[0].decode()  # type: ignore
    assert f"'{vstr}'" in cur._query.query.decode()
    assert f"'{vstr}'" in cur._query.params[1].decode()  # type: ignore


def test_mogrify(conn):
    cur = psycopg.ClientCursor(conn)
    res = cur.mogrify(t"select {vint}, {vstr} as {vstr:i}")
    assert res == "select 16, 'hello' as \"hello\""


def test_raw_cursor(conn):
    cur = psycopg.RawCursor(conn)
    with pytest.raises(psycopg.NotSupportedError):
        cur.execute(t"select {vint}, {vstr} as {vstr:i}")


def test_server_cursor(conn):
    with psycopg.ServerCursor(conn, "test") as cur:
        cur.execute(t"select {vint}, {vstr} as {vstr:i}")
        assert cur.fetchone() == (vint, vstr)
        assert cur.description
        assert cur.description[1].name == vstr
        assert cur._query
        assert b"$2" in cur._query.query
        assert b"$3" not in cur._query.query


@skip_async
def test_as_string():
    query = sql.as_string(t"select {vstr}")
    assert query == no_e("select 'hello'")


def test_as_string_context(conn):
    conn.adapters.register_dumper(str, make_dumper("1"))
    query = sql.as_string(t"select {vstr}", context=conn)
    assert query == no_e("select 'hello1'")


@skip_async
def test_as_bytes():
    query = sql.as_bytes(t"select {vstr}")
    assert query == no_e(b"select 'hello'")


def test_as_bytes_context(conn):
    conn.adapters.register_dumper(str, make_dumper("1"))
    query = sql.as_bytes(t"select {vstr}", context=conn)
    assert query == no_e(b"select 'hello1'")
